/*
    Fichier cylincr.c
    Auteur Bernard Chardonneau

    Logiciel libre, droits d'utilisation prciss en franais
    dans le fichier : licence-fr.txt

    Traductions des droits d'utilisation dans les fichiers :
    licence-en.txt , licence-es.txt , licence-pt.txt ,
    licence-eo.txt , licence-eo-utf.txt

    Droits d'utilisation galement sur la page web :
    http://cyloop.tuxfamily.org/voir.php?page=droits


    Programme permettant d'incrmenter la variable d'un fichier
    cyloop si elle est de type compteur.
*/

#include <stdio.h>
#include <time.h>
#include "types.h"
#include "limites.h"
#include "messages.h"
#include "majcycle.h"


// prototypes
void pondere (int coef_pond);
void incrval (uint16 num_instant, int majcmoy);



// variables globales (pour viter des tas de passages de paramtre)
FILE        *descfic;  // descripteur du fichier cyloop
desc_cyloop entetefic; // entte du fichier cyloop
desc_var    entetevar; // entte de la variable du fichier cyloop



// programme principal

main (int narg, char *varg[])
{
    char      *nomfic;       // nom du fichier cyloop
    struct tm stm;           // structure contenant la date et l'heure courante
    uint32    timep;         // date et heure dans un uint32
    int       maj_entetefic; // indicateur pour mise  jour fichier
    uint16    num_instant;     // numro de la donne  incrmenter
    uint32    duree_cycle;   // dure du cycle


    // mmorisation du nom de la commande
    memcom (*varg);

    // rcupration du nom de fichier et rajout ventuel du suffixe
    if (--narg)
        nomfic = recup_nomfic (varg [1]);
    else
    {
        // "Nom de fichier manquant"
        // "Syntaxe : %s nom_fichier_cyloop"
        affiche_err ("MANQUE_NOMFICH");
        psyntaxe ("SYNT_CYLINCR");
    }

    // ouverture en lecture / criture du fichier cyloop
    descfic = fopen (nomfic, "r+");

    if (! descfic)
        // "Fichier_cyloop %s inexistant ou protg en lecture ou criture"
        err_fatale_arg ("ACCES_CYLOOP_RW",
                                                                   nomfic);

    // lecture de l'entte du fichier cyloop
    fread (&entetefic, 16, 1, descfic);

    // vrification de la signature
    if (entetefic.signature [0] != 'c' || entetefic.signature [1] != 'y'
                                       || entetefic.signature [2] != 'l')
    {
        // "Le fichier %s n'est pas un fichier_cyloop"
        aff_err_arg ("FICH_NON_CYLOOP", nomfic);
        fclose (descfic);
        return;
    }

    // vrification de la valeur nb_var
    if (entetefic.nb_var)
    {
        // "Ce fichier cyloop utilise une liste de variables"
        // "utiliser la version complte de cyloop"
        affiche_err ("CYLOOP_LISTEVAR");
        affiche_err ("UTIL_VCOMPLETE");

        fclose (descfic);
        return;
    }

    // lecture de la partie fixe de l'entte de la variable du fichier cyloop
    fread (&entetevar, 6, 1, descfic);

    // vrification de la nature de la variable
    if (! (entetevar.typedon & compteur))
    {
   // "La variable du fichier cyloop n'est pas un compteur : utiliser cyladdval"
        affiche_err ("VAR_NON_COMPTEUR");

        fclose (descfic);
        return;
    }

    // fin de lecture de l'entte de la variable si ncessaire
    if (! (entetevar.typedon & cmoy_sep))
        fread (&entetevar.coef_moy, 2, 1, descfic);

    // rcuprer la date et l'heure courante
    time (&timep);

    // si le dbut du cycle est ultrieur  l'instant prsent
    if (timep < entetefic.deb_cycle)
    {
        fclose (descfic);

        // "Rgression dans le temps : une intervention manuelle "
        // "sur le fichier cyloop ou l'horloge est ncessaire"
        affiche_err ("REGRES_TEMPS");
        err_fatale ("REGRES_TEMPS2");
    }

    // initialisation indicateurs de mise  jour du fichier
    maj_entetefic = NON;

    // si on utilise l'heure locale
    if (! (entetefic.caract_gen & temps_utc))
    {
        // rcuprer le dtail de la date et de l'heure locale
        localtime_r (&timep, &stm);

        // si changement d'heure t/hiver
        if ((entetefic.etat_cyloop & heure_ete) != stm.tm_isdst)
            // prendre en compte ce changement
            maj_entetefic = maj_type_heure (stm.tm_isdst, timep);
    }

    // si changement de cycle
    if (entetefic.deb_cycle + entetefic.duree_cycle <= timep)
    {
        // prise en compte de ce changement de cycle
        changement_cycle (timep);

        // une manire de mmoriser un dbut de cycle
        entetevar.der_enreg = 0;

        // on reportera la mise  jour dans le fichier cyloop
        maj_entetefic = OUI;
    }

    // si mise  jour de l'entte du fichier cyloop ncessaire
    if (maj_entetefic)
    {
        // recopier la nouvelle entte
        rewind (descfic);
        fwrite (&entetefic, 16, 1, descfic);
    }

    // rcupration de la dure du cycle actuel
    duree_cycle = duree_cycle_actuel (&timep);

    // calcul de l'instant du cycle  mettre  jour
    num_instant = calcul_instant (timep, duree_cycle);

    // si on a chang de cycle et une seule valeur coef_moy pour la variable
    if (! entetevar.der_enreg && !(entetevar.typedon & cmoy_sep))
    {
        // mettre  jour cette valeur
        if (entetevar.ponderation)
            entetevar.coef_moy += 256;
        else
            entetevar.coef_moy ++;

        // y compris dans le fichier cyloop
        fseek (descfic, 22, SEEK_SET);
        fwrite (&entetevar.coef_moy, 2, 1, descfic);
    }

    // si c'est la premire mise  jour effectue durant ce cycle
    // ou si on passe  un autre instant du mme cycle
    if (entetevar.der_enreg < num_instant)
    {
        // mettre  jour le numro de l'instant du cycle trait
        entetevar.der_enreg = num_instant;

        // y compris dans le fichier cyloop
        fseek (descfic, 16, SEEK_SET);
        fwrite (&entetevar.der_enreg, 2, 1, descfic);

        // incrmenter le compteur et si ncessaire le
        // champ coef_moy pour l'instant du cycle concern
        incrval (num_instant, OUI);
    }
    // sinon, si on a dj mis  jour le compteur pour cet instant du cycle
    else if (entetevar.der_enreg == num_instant)
    {
        // si on est autoris  le mettre  jour plusieur
        // fois pour chacun des instants du cycle
        if (entetevar.typedon & cumul_donnees)
            // incrmenter le compteur et si ncessaire le
            // champ coef_moy pour l'instant du cycle concern
            incrval (num_instant, entetevar.typedon & cpt_toute_don);
    }
    // sinon, il y a rgression dans le temps
    else
    {
        // vrifier si elle est suprieure  2 heures
        if (calcul_instant (timep + 7200, duree_cycle) < entetevar.der_enreg)
        {
            // vrifier si elle est suprieure  1 journe
            // et afficher le message d'erreur appropri
            if (calcul_instant (timep + secjour, duree_cycle) <
                                                 entetevar.der_enreg)
            {
                // "Forte rgression dans le temps : une intervention manuelle "
                // "sur le fichier cyloop ou l'horloge est ncessaire"
                affiche_err ("FORTE_REGR_TEMPS");
                affiche_err ("REGRES_TEMPS2");
            }
            else
              // "Rgression dans le temps : vrifier l'horloge de l'ordinateur"
                affiche_err ("REGR_VERIF_HORL");
        }
    }

    // le fichier cyloop a t mis  jour
    fclose (descfic);
}



// calcul de pondration sur les cycles antrieurs

void pondere (int coef_pond)
{
    uint16 coef_moy; // coef_moy si distinct pour chaque instant du cycle
    uint16 valeur16b;  // valeur  pondrer si elle est sur 16 bits
    uint32 valeur32b;  // valeur  pondrer si elle est sur 32 bits
    uint32 valmaj;     // valeur sur 32 bits pour calcul de pondration
    int    i;          // compteur


    // si coefficient coef_moy global au compteur
    if (! (entetevar.typedon & cmoy_sep))
    {
        // mettre  jour ce coefficient
        // le calcul de multiplication et division est fait sur 32 bits
        valmaj = entetevar.coef_moy;
        entetevar.coef_moy = (valmaj * coef_pond) >> 8;

        // ce coefficient sera mis  jour plus tard dans le fichier

        // se positionner sur la premire valeur du compteur
        fseek (descfic, 24, SEEK_SET);

        // si compteur sur 32 bits
        if (entetevar.typedon & donnees_32bits)
        {
            // mise  jour des valeurs du compteur 32 bits
            for (i = entetefic.nb_don_cycle; i > 0; i--)
            {
                // resynchronisation position dans le fichier
                // fseek (descfic, 0, SEEK_CUR); (vrifier utilit)

                // rcupration de la valeur du compteur sur 32 bits
                fread (&valeur32b, 4, 1, descfic);

                // si risque de dbordement lors du calcul
                if (valeur32b & 0xFF000000)
                    // faire la division d'abord
                    valeur32b = (valeur32b >> 8) * coef_pond;

                // sinon faire la multiplication d'abord
                else
                    valeur32b = (valeur32b * coef_pond) >> 8;

                // mettre  jour la valeur du compteur dans le fichier
                fseek (descfic, -4, SEEK_CUR);
                fwrite (&valeur32b, 4, 1, descfic);
            }
        }
        // sinon (compteur sur 16 bits)
        else
        {
            // mise  jour des valeurs du compteur 16 bits
            for (i = entetefic.nb_don_cycle; i > 0; i--)
            {
                // resynchronisation position dans le fichier
                // fseek (descfic, 0, SEEK_CUR); (vrifier utilit)

                // rcupration de la valeur du compteur sur 16 bits
                fread (&valeur16b, 2, 1, descfic);

                // on fait le calcul de pondration sur 32 bits
                valmaj = valeur16b;
                valeur16b = (valmaj * coef_pond) >> 8;

                // mettre  jour la valeur du compteur dans le fichier
                fseek (descfic, -2, SEEK_CUR);
                fwrite (&valeur16b, 2, 1, descfic);
            }
        }
    }
    // sinon
    else
    {
        // se positionner sur le premier coefficient coef_moy de la variable
        fseek (descfic, 22, SEEK_SET);

        // si compteur sur 32 bits
        if (entetevar.typedon & donnees_32bits)
        {
            // mise  jour des valeurs de coef_moy et du compteur 32 bits
            for (i = entetefic.nb_don_cycle; i > 0; i--)
            {
                // resynchronisation position dans le fichier
                // fseek (descfic, 0, SEEK_CUR); (vrifier utilit)

                // rcupration de la valeur du coef_moy
                fread (&coef_moy, 2, 1, descfic);

                // rcupration de la valeur du compteur sur 32 bits
                fread (&valeur32b, 4, 1, descfic);

                // calcul de pondration pour coef_moy
                // le calcul de multiplication et division est fait sur 32 bits
                valmaj = coef_moy;
                coef_moy = (valmaj * coef_pond) >> 8;

                // calcul de pondration pour la valeur du compteur
                // si risque de dbordement lors du calcul
                if (valeur32b & 0xFF000000)
                    // faire la division d'abord
                    valeur32b = (valeur32b >> 8) * coef_pond;

                // sinon faire la multiplication d'abord
                else
                    valeur32b = (valeur32b * coef_pond) >> 8;

                // mettre  jour les valeurs dans le fichier
                fseek (descfic, -6, SEEK_CUR);
                fwrite (&coef_moy, 2, 1, descfic);
                fwrite (&valeur32b, 4, 1, descfic);
            }
        }
        // sinon (compteur sur 16 bits)
        else
        {
            // mise  jour des valeurs de coef_moy et du compteur 16 bits
            // la mme boucle sert aux 2 valeurs, moins optimis pour les
            // accs au fichier, mais plus simple
            for (i = entetefic.nb_don_cycle * 2; i > 0; i--)
            {
                // resynchronisation position dans le fichier
                // fseek (descfic, 0, SEEK_CUR); (vrifier utilit)

                // rcupration de la valeur du compteur sur 16 bits
                fread (&valeur16b, 2, 1, descfic);

                // on fait le calcul de pondration sur 32 bits
                valmaj = valeur16b;
                valeur16b = (valmaj * coef_pond) >> 8;

                // mettre  jour la valeur du compteur dans le fichier
                fseek (descfic, -2, SEEK_CUR);
                fwrite (&valeur16b, 2, 1, descfic);
            }
        }
    }
}



// incrmente le compteur pour l'instant num_instant du cycle

void incrval (uint16 num_instant, int majcmoy)
{
    int    ajout;      // valeur  rajouter au compteur
    long   posval;     // position de la valeur  incrmenter dans le fichier
    uint16 coef_moy; // coefficient coef_moy pour la valeur  incrmenter
    uint16 valeur16b;  // valeur  incrmenter si elle est sur 16 bits
    uint32 valeur32b;  // valeur  incrmenter si elle est sur 32 bits
    int    majfich;    // indique si on devra mettre  jour le fichier


    // initialisation
    majfich = OUI;

    // calcul de la valeur  rajouter au compteur
    if (entetevar.ponderation)
        ajout = 256;
    else
        ajout = 1;

    // pour le calcul de position, le premier instant
    // du cycle n'entrainera pas de dcalage
    num_instant --;

    // calcul de la position dans le fichier de la valeur  incrmenter
    posval = 24;

    if (entetevar.typedon & cmoy_sep)
        posval += num_instant * 2;

    if (entetevar.typedon & donnees_32bits)
        posval += num_instant * 4;
    else
        posval += num_instant * 2;

    // si coefficient coef_moy  mettre  jour
    if (majcmoy && (entetevar.typedon & cmoy_sep))
    {
        // se positionner dans le fichier sur la zone  modifier
        fseek (descfic, posval - 2, SEEK_SET);

        // rcuprer et incrmenter le coefficient coef_moy
        fread (&coef_moy, 2, 1, descfic);
        coef_moy += ajout;
    }
    // sinon, juste valeur du compteur  modifier
    else
        // se positionner dans le fichier sur la zone  modifier
        fseek (descfic, posval, SEEK_SET);

    // si le compteur est sur 32 bits
    if (entetevar.typedon & donnees_32bits)
    {
        // rcupration de la valeur du compteur sur 32 bits
        fread (&valeur32b, 4, 1, descfic);

        // incrementation de la valeur du compteur
        valeur32b += ajout;
    }
    // sinon (compteur sur 16 bits)
    else
    {
        // rcupration de la valeur du compteur sur 16 bits
        fread (&valeur16b, 2, 1, descfic);

        // s'il n'y a pas encore eu de dbordement numrique
        if (valeur16b != max_uint16)
        {
            // incrmenter la valeur (calcul sur 32 bits)
            valeur32b = (uint32) valeur16b + ajout;

            // si dbordement numrique
            if (valeur32b & 0xFFFF0000)
            {
                // message d'erreur
                // "Dbordement numrique pour la variable"
                // "Il faudrait la convertir en 32 bits"
                affiche_err ("DEBORD_NUMERIQUE");
                affiche_err ("BESOIN_CONV32B");

                // le compteur prendra la valeur maximale
                valeur16b = max_uint16;
            }
            // sinon, mmorisation de la nouvelle valeur du compteur
            else
                valeur16b = valeur32b;
        }
        // sinon, message de rappel dbordement numrique
        else
        {
            // "Rappel : Dbordement numrique pour la variable"
            // "Il faudrait la convertir en 32 bits"
            affiche_err ("PAR_DEBORD_NUM");
            affiche_err ("BESOIN_CONV32B");

            // pas la peine de mettre  jour le fichier pour la variable
            majfich = NON;
        }
    }

    // si coefficient coef_moy  mettre  jour
    if (majcmoy && (entetevar.typedon & cmoy_sep))
    {
        // se positionner dans le fichier sur la zone  modifier
        fseek (descfic, posval - 2, SEEK_SET);

        // mettre  jour le coefficient coef_moy
        fwrite (&coef_moy, 2, 1, descfic);
    }
    // sinon si juste mise  jour de la variable
    else if (majfich)
    {
        // se positionner dans le fichier sur la zone  modifier
        fseek (descfic, posval, SEEK_SET);
    }

    // mettre  jour la valeur de la variable (en fonction de sa taille)
    if (majfich)
    {
        if (entetevar.typedon & donnees_32bits)
            fwrite (&valeur32b, 4, 1, descfic);
        else
            fwrite (&valeur16b, 2, 1, descfic);
    }
}
