/*
    Fichier dialogue.c
    Auteur Bernard Chardonneau

    Logiciel libre, droits d'utilisation prciss en franais
    dans le fichier : licence-fr.txt

    Traductions des droits d'utilisation dans les fichiers :
    licence-en.txt , licence-es.txt , licence-pt.txt ,
    licence-eo.txt , licence-eo-utf.txt

    Droits d'utilisation galement sur la page web :
    http://cyloop.tuxfamily.org/voir.php?page=droits


    Bibliothque de fonctions utilise par le progranne cylcree
    pour rcuprer les choix de l'utilisateur.
*/

#include <stdio.h>
#include <stdarg.h>
#include "messages.h"
#include "dialogue.h"


// saisie d'une rponse sur un caractre au clavier
// vite les problmes de fonctionnent de scanf

char lirep ()
{
    int  caract;
    char reponse;

    // on rcupre 1 caractre
    caract = getchar ();

    // on le passe en minuscules
    reponse = caract | 0x20;

    // on saute tous les autres caractres de la ligne
    while (caract != '\n')
        caract = getchar ();

    // et on retourne le caractre mmoris
    return (reponse);
}



// saisie d'un chiffre au clavier

int lecchiffre ()
{
    char valeur;

    // on rcupre 1 caractre
    valeur = lirep ();

    // si c'est un chiffre, on retourne sa valeur
    if ('0' < valeur && valeur <= '9')
        return (valeur & 0x0F);
    else
        // sinon, on retourne 0
        return (0);
}



// saisie d'un entier au clavier sans utiliser scanf

int lirentier ()
{
    int  valeur;
    char caract;

    // initialisation
    valeur = 0;
    caract = getchar ();

    // tant que des chiffres sont saisis au clavier
    while ('0' <= caract && caract <= '9')
    {
        // prendre en compte le dernier chiffre lu
        valeur = (valeur * 10) + caract - '0';

        // et lire le caractre suivant
        caract = getchar ();
    }

    // sauter tous les autres caractres de la ligne
    while (caract != '\n')
        caract = getchar ();

    // et retourner la valeur saisie
    return (valeur);
}



// saisie et traitement du coefficient de pondration

int liponder ()
{
    float valeur;
    float puis10;
    int   diviseur;
    char  caract;

    // initialisation
    valeur = 0;
    caract = getchar ();

    // tant que des chiffres sont saisis au clavier
    while ('0' <= caract && caract <= '9')
    {
        // prendre en compte le dernier chiffre lu
        valeur = (valeur * 10) + caract - '0';

        // et lire le caractre suivant
        caract = getchar ();
    }

    // si partie dcimale
    if (caract == '.' || caract == ',')
    {
        puis10 = 0.1;
        caract = getchar ();

        while ('0' <= caract && caract <= '9')
        {
            // prendre en compte le dernier chiffre lu
            valeur = valeur + (caract - '0') * puis10;

            // et lire le caractre suivant
            caract = getchar ();
            puis10 = puis10 / 10;
        }
    }

    // sauter les blancs ventuels aprs le nombre
    while (caract == ' ')
        caract = getchar ();

    if (caract == '%')
        valeur = valeur * 2.56;
    else if (caract == '/')
    {
        do
            caract = getchar ();
        while (caract == ' ');

        diviseur = 0;

        while ('0' <= caract && caract <= '9')
        {
            // prendre en compte le dernier chiffre lu
            diviseur = (diviseur * 10) + caract - '0';

            // et lire le caractre suivant
            caract = getchar ();
        }

        if (diviseur != 256 && diviseur != 0)
            valeur = valeur * 256 / diviseur;
    }
    else if (caract == '-')
        valeur = -1;  // ce qui sera dtect comme une erreur
    else
        valeur = valeur * 256;

    // sauter tous les autres caractres de la ligne
    while (caract != '\n')
        caract = getchar ();

    // et retourner la valeur saisie
    return ((int) (valeur + 0.5));
}



// saisie d'un choix parmi une liste de propositions

int lecchoix (char *chaine1, ...)
{
    char *listechaines [10];
    char reponse [4];
    va_list ap;
    int  nb_choix;
    int  choix;
    int  i;

    // mmorisation du libell du premier choix
    listechaines [0] = chaine1;

    // initialisation boucle de rcupration des libells des choix
    va_start (ap, chaine1);
    nb_choix = 0;

    // mmorisation et comptage des libells des diffrents choix
    do
        listechaines [++nb_choix] = va_arg (ap, char *);
    while (nb_choix < 9 && listechaines [nb_choix]);

    // rcupration termine
    va_end (ap);

    if (listechaines [nb_choix])
        // "Fonction lecchoix : liste de choix mal termine"
        err_fatale ("ERR_LECCHOIX");

    // boucle d'affichage et de saisie des choix, jusqu' un choix correct
    do
    {
        // affichage des diffrents choix
        for (i = 0; i < nb_choix; i++)
            printf ("%d : %s\n", i + 1, message (listechaines [i]));

        // "\nVotre choix ? "
        affiche_msg ("VOTRE_CHOIX");

        // saisie du choix par l'oprateur
        choix = lecchiffre ();

        // vrification du choix saisi
        if (choix < 1 || choix > nb_choix)
            // "Choix incorrect !\n"
            affiche_err ("MAUV_CHOIX");
    }
    // on recommence si le choix tait incorrect
    while (choix < 1 || choix > nb_choix);

    // choix valid
    return choix;
}
